"""T-Copula."""
import numpy
from scipy import special
import chaospy

from ..baseclass import CopulaDistribution, Distribution


class t_copula(Distribution):
    """T-Copula."""

    def __init__(self, df, covariance, rotation=None):
        assert isinstance(df, float)
        covariance = numpy.asarray(covariance)
        assert covariance.ndim == 2, "Covariance must be a matrix"
        assert covariance.shape[0] == covariance.shape[1], (
            "Parameters 'covariance' not a square matrix.")

        dependencies, _, rotation = chaospy.declare_dependencies(
            self,
            parameters=dict(covariance=covariance),
            rotation=rotation,
            dependency_type="accumulate",
        )
        correlation = covariance/numpy.sqrt(numpy.outer(
            numpy.diag(covariance), numpy.diag(covariance)))
        self._permute = numpy.eye(len(rotation), dtype=int)[rotation]
        self._correlation = self._permute.dot(correlation).dot(self._permute.T)
        cholesky = numpy.linalg.cholesky(self._correlation)
        self._fwd_transform = self._permute.T.dot(numpy.linalg.inv(cholesky))
        self._inv_transform = self._permute.T.dot(cholesky)

        super(t_copula, self).__init__(
            parameters=dict(df=df),
            dependencies=dependencies,
            rotation=rotation,
            repr_args=[covariance.tolist()],
        )

    def _cdf(self, xloc, idx, df, cache):
        dim = self._rotation.index(idx)
        conditions = [self._get_cache(dim_, cache, get=0)
                      for dim_ in self._rotation[:dim]]
        assert not any([isinstance(condition, chaospy.Distribution)
                        for condition in conditions])
        xloc = numpy.vstack(conditions+[xloc])
        zloc = self._fwd_transform[idx, :len(xloc)].dot(special.stdtrit(df, xloc))
        out = special.stdtr(df, zloc)
        return out

    def _ppf(self, qloc, idx, df, cache):
        dim = self._rotation.index(idx)
        conditions = [self._get_cache(dim_, cache, get=1)
                      for dim_ in self._rotation[:dim]]
        assert not any([isinstance(condition, chaospy.Distribution)
                        for condition in conditions])
        qloc = numpy.vstack(conditions+[qloc])
        zloc = special.stdtrit(df, qloc)
        out = special.stdtr(df, self._inv_transform[idx, :len(qloc)].dot(zloc))
        return out

    def _pdf(self, xloc, idx, df, cache):
        raise chaospy.UnsupportedFeature("Copula not supported.")

    def _lower(self, idx, df, cache):
        return 0.

    def _upper(self, idx, df, cache):
        return 1.


class TCopula(CopulaDistribution):
    """
    T-Copula.

    Args:
        dist (Distribution):
            The distribution to wrap in a copula.
        R (numpy.ndarray):
            Covariance matrix defining dependencies..
        df (float):
            The degree of freedom in the underlying student-t distribution.

    Examples:
        >>> distribution = chaospy.TCopula(
        ...     chaospy.Iid(chaospy.Uniform(-1, 1), 2),
        ...     df=5, covariance=[[1, .5], [.5, 1]])
        >>> distribution
        TCopula(Iid(Uniform(lower=-1, upper=1), 2), 5.0, [[1.0, 0.5], [0.5, 1.0]])
        >>> samples = distribution.sample(3)
        >>> samples.round(4)
        array([[ 0.3072, -0.77  ,  0.9006],
               [ 0.1274,  0.3147,  0.1928]])
        >>> distribution.pdf(samples).round(4)
        array([0.2932, 0.1367, 0.1969])
        >>> distribution.fwd(samples).round(4)
        array([[0.6536, 0.115 , 0.9503],
               [0.4822, 0.8725, 0.2123]])
        >>> mesh = numpy.meshgrid([.4, .5, .6], [.4, .5, .6])
        >>> distribution.inv(mesh).round(4)
        array([[[-0.2   ,  0.    ,  0.2   ],
                [-0.2   ,  0.    ,  0.2   ],
                [-0.2   ,  0.    ,  0.2   ]],
        <BLANKLINE>
               [[-0.2699, -0.1738, -0.0741],
                [-0.1011,  0.    ,  0.1011],
                [ 0.0741,  0.1738,  0.2699]]])

    """

    def __init__(self, dist, df, covariance):
        assert len(dist) == len(covariance)
        df = float(df)
        covariance = numpy.asfarray(covariance)
        super(TCopula, self).__init__(
            dist=dist,
            trans=t_copula(df, covariance, dist._rotation),
            repr_args=[dist, df, covariance.tolist()],
        )



"""
Clenshaw-Curtis quadrature method is a good all-around quadrature method
comparable to Gaussian quadrature, but typically limited to finite intervals
without a specific weight function. In addition to be quite accurate, the
weights and abscissas can be calculated quite fast.

Another thing to note is that Clenshaw-Curtis, with an appropriate growth rule
is fully nested. This means, if one applies a method that combines different
order of quadrature rules, the number of evaluations can often be reduced as
the abscissas can be used across levels.

Example usage
-------------

The first few orders with linear growth rule::

    >>> distribution = chaospy.Uniform(0, 1)
    >>> for order in [0, 1, 2, 3]:
    ...     abscissas, weights = chaospy.generate_quadrature(
    ...         order, distribution, rule="clenshaw_curtis")
    ...     print(order, abscissas.round(3), weights.round(3))
    0 [[0.5]] [1.]
    1 [[0. 1.]] [0.5 0.5]
    2 [[0.  0.5 1. ]] [0.167 0.667 0.167]
    3 [[0.   0.25 0.75 1.  ]] [0.056 0.444 0.444 0.056]

The first few orders with exponential growth rule::

    >>> for order in [0, 1, 2]:
    ...     abscissas, weights = chaospy.generate_quadrature(
    ...         order, distribution, rule="clenshaw_curtis", growth=True)
    ...     print(order, abscissas.round(3), weights.round(3))
    0 [[0.5]] [1.]
    1 [[0.  0.5 1. ]] [0.167 0.667 0.167]
    2 [[0.    0.146 0.5   0.854 1.   ]] [0.033 0.267 0.4   0.267 0.033]

Applying the rule using Smolyak sparse grid::

    >>> distribution = chaospy.Iid(chaospy.Uniform(0, 1), 2)
    >>> abscissas, weights = chaospy.generate_quadrature(
    ...     2, distribution, rule="clenshaw_curtis",
    ...     growth=True, sparse=True)
    >>> abscissas.round(2)
    array([[0.  , 0.  , 0.  , 0.15, 0.5 , 0.5 , 0.5 , 0.5 , 0.5 , 0.85, 1.  ,
            1.  , 1.  ],
           [0.  , 0.5 , 1.  , 0.5 , 0.  , 0.15, 0.5 , 0.85, 1.  , 0.5 , 0.  ,
            0.5 , 1.  ]])
    >>> weights.round(3)
    array([ 0.028, -0.022,  0.028,  0.267, -0.022,  0.267, -0.089,  0.267,
           -0.022,  0.267,  0.028, -0.022,  0.028])
"""
from __future__ import division
try:
    from functools import lru_cache
except ImportError:
    from functools32 import lru_cache

import numpy
import chaospy

from .combine import combine_quadrature


def quad_clenshaw_curtis(order, domain, growth=False, segments=1):
    """
    Generate the quadrature nodes and weights in Clenshaw-Curtis quadrature.

    Args:
        order (int, numpy.ndarray):
            Quadrature order.
        domain (chaospy.distributions.baseclass.Distribution, numpy.ndarray):
            Either distribution or bounding of interval to integrate over.
        growth (bool):
            If True sets the growth rule for the quadrature rule to only
            include orders that enhances nested samples.
        segments (int):
            Split intervals into N subintervals and create a patched
            quadrature based on the segmented quadrature. Can not be lower than
            `order`. If 0 is provided, default to square root of `order`.
            Nested samples only exist when the number of segments are fixed.

    Returns:
        (numpy.ndarray, numpy.ndarray):
            abscissas:
                The quadrature points for where to evaluate the model function
                with ``abscissas.shape == (len(dist), N)`` where ``N`` is the
                number of samples.
            weights:
                The quadrature weights with ``weights.shape == (N,)``.

    Example:
        >>> abscissas, weights = quad_clenshaw_curtis(4, (0, 1))
        >>> abscissas.round(4)
        array([[0.    , 0.1464, 0.5   , 0.8536, 1.    ]])
        >>> weights.round(4)
        array([0.0333, 0.2667, 0.4   , 0.2667, 0.0333])
        >>> abscissas, weights = quad_clenshaw_curtis(4, (0, 1), segments=0)
        >>> abscissas.round(4)
        array([[0.  , 0.25, 0.5 , 0.75, 1.  ]])
        >>> weights.round(4)
        array([0.0833, 0.3333, 0.1667, 0.3333, 0.0833])
    """
    if isinstance(domain, chaospy.Distribution):
        abscissas, weights = quad_clenshaw_curtis(
            order, (domain.lower, domain.upper), growth, segments)

        # Sometimes edge samples (inside the domain) falls out again from simple
        # rounding errors. Edge samples needs to be adjusted.
        eps = 1e-14*(domain.upper-domain.lower)
        abscissas_ = numpy.clip(abscissas.T, domain.lower+eps, domain.upper-eps).T
        weights *= domain.pdf(abscissas_).flatten()
        weights /= numpy.sum(weights)

        return abscissas, weights

    order = numpy.asarray(order, dtype=int).flatten()
    lower, upper = numpy.array(domain)
    lower = numpy.asarray(lower).flatten()
    upper = numpy.asarray(upper).flatten()

    dim = max(lower.size, upper.size, order.size)

    order = order*numpy.ones(dim, dtype=int)
    lower = lower*numpy.ones(dim)
    upper = upper*numpy.ones(dim)
    segments = segments*numpy.ones(dim, dtype=int)

    if growth:
        order = numpy.where(order > 0, 2**order, 0)

    abscissas, weights = zip(*[_clenshaw_curtis(order_, segment)
                               for order_, segment in zip(order, segments)])

    return combine_quadrature(abscissas, weights, (lower, upper))


@lru_cache(None)
def _clenshaw_curtis(order, segments=1):
    r"""
    Backend method.

    Examples:
        >>> abscissas, weights = _clenshaw_curtis(0, 0)
        >>> abscissas
        array([0.5])
        >>> weights
        array([1.])
        >>> abscissas, weights = _clenshaw_curtis(2, 0)
        >>> abscissas
        array([0. , 0.5, 1. ])
        >>> weights
        array([0.16666667, 0.66666667, 0.16666667])
        >>> abscissas, weights = _clenshaw_curtis(4, 0)
        >>> abscissas
        array([0.  , 0.25, 0.5 , 0.75, 1.  ])
        >>> weights
        array([0.08333333, 0.33333333, 0.16666667, 0.33333333, 0.08333333])
        >>> abscissas, weights = _clenshaw_curtis(8, 0)
        >>> abscissas.round(3)
        array([0.   , 0.073, 0.25 , 0.427, 0.5  , 0.573, 0.75 , 0.927, 1.   ])
        >>> weights.round(3)
        array([0.017, 0.133, 0.2  , 0.133, 0.033, 0.133, 0.2  , 0.133, 0.017])
        >>> abscissas, weights = _clenshaw_curtis(16, 0)
        >>> abscissas.round(3)
        array([0.   , 0.037, 0.125, 0.213, 0.25 , 0.287, 0.375, 0.463, 0.5  ,
               0.537, 0.625, 0.713, 0.75 , 0.787, 0.875, 0.963, 1.   ])
        >>> weights.round(3)
        array([0.008, 0.067, 0.1  , 0.067, 0.017, 0.067, 0.1  , 0.067, 0.017,
               0.067, 0.1  , 0.067, 0.017, 0.067, 0.1  , 0.067, 0.008])
    """
    if segments != 1 and order > 2:
        if not segments:
            segments = int(numpy.sqrt(order))
        assert segments < order, "few samples to distribute than intervals"
        abscissas = []
        weights = []

        nodes = numpy.linspace(0, 1, segments+1)
        for idx, (lower, upper) in enumerate(zip(nodes[:-1], nodes[1:])):

            order_ = order//segments + (idx < (order%segments))
            abscissa, weight = _clenshaw_curtis(order_, segments=1)
            abscissa = abscissa*(upper-lower) + lower
            weight = weight*(upper-lower)
            if abscissas:
                weights[-1] += weight[0]
                abscissa = abscissa[1:]
                weight = weight[1:]
            abscissas.extend(abscissa)
            weights.extend(weight)

        assert len(abscissas) == order+1, (len(abscissas), order+1)
        assert len(weights) == order+1
        return numpy.array(abscissas), numpy.array(weights)

    if order == 0:
        return numpy.array([.5]), numpy.array([1.])

    theta = (order-numpy.arange(order+1))*numpy.pi/order
    abscissas = 0.5*numpy.cos(theta) + 0.5

    idx, idy = numpy.mgrid[:order+1, :order//2]
    weights = 2*numpy.cos(2*(idy+1)*theta[idx])/(4*idy*(idy+2)+3)
    if order % 2 == 0:
        weights[:, -1] *= 0.5
    weights = (1-numpy.sum(weights, -1)) / order

    weights[0] /= 2
    weights[-1] /= 2

    assert len(abscissas) == order+1
    return abscissas, weights



r"""
Copulas are a type dependency structure imposed on independent variables to
achieve to more complex problems without adding too much complexity.

To construct a copula one needs a copula transformation and the
Copula wrapper::

    >>> dist = chaospy.Iid(chaospy.Uniform(), 2)
    >>> copula = chaospy.Gumbel(dist, theta=1.5)

The resulting copula is then ready for use::

    >>> copula.sample(5).round(4)
    array([[0.6536, 0.115 , 0.9503, 0.4822, 0.8725],
           [0.6286, 0.0654, 0.96  , 0.5073, 0.9705]])

"""
from .archimedean import Archimedean

from .gumbel import Gumbel
from .clayton import Clayton
from .joe import Joe
from .nataf import Nataf
from .t_copula import TCopula



"""Burr Type III distribution."""
import numpy
from scipy import special

from ..baseclass import SimpleDistribution, ShiftScaleDistribution


class burr(SimpleDistribution):
    """Stadard Burr distribution."""

    def __init__(self, alpha=1., kappa=1.):
        super(burr, self).__init__(dict(alpha=alpha, kappa=kappa))

    def _pdf(self, x, alpha, kappa):
        output = numpy.zeros(x.shape)
        indices = x > 0
        output[indices] = alpha*kappa*x[indices]**(-alpha-1.0)
        output[indices] /= (1+x[indices]**alpha)**(kappa+1)
        return output

    def _cdf(self, x, alpha, kappa):
        output = numpy.zeros(x.shape)
        indices = x > 0
        output[indices] = 1-(1+x[indices]**alpha)**-kappa
        return output

    def _ppf(self, q, alpha, kappa):
        return ((1-q)**(-1./kappa) -1)**(1./alpha)

    def _mom(self, k, alpha, kappa):
        return kappa*special.beta(1-k*1./alpha, kappa+k*1./alpha)

    def _lower(self, alpha, kappa):
        del alpha
        del kappa
        return 0.

    def _upper(self, alpha, kappa):
        return (1e12**(1./kappa)-1)**(1./alpha)


class Burr(ShiftScaleDistribution):
    """
    Burr Type XII or Singh-Maddala distribution.

    Args:
        alpha (float, Distribution):
            First shape parameter
        kappa (float, Distribution):
            Second shape parameter
        loc (float, Distribution):
            Location parameter
        scale (float, Distribution):
            Scaling parameter

    Examples:
        >>> distribution = chaospy.Burr(5, 2)
        >>> distribution
        Burr(5, 2)
        >>> uloc = numpy.linspace(0, 1, 6)
        >>> uloc
        array([0. , 0.2, 0.4, 0.6, 0.8, 1. ])
        >>> xloc = distribution.inv(uloc)
        >>> xloc.round(3)
        array([ 0.   ,  0.652,  0.781,  0.897,  1.043, 15.849])
        >>> numpy.allclose(distribution.fwd(xloc), uloc)
        True
        >>> distribution.pdf(xloc).round(3)
        array([ 0.   , 92.945, 20.444,  4.852,  0.694,  0.   ])
        >>> distribution.sample(4).round(3)
        array([0.931, 0.575, 1.284, 0.828])
        >>> distribution.mom(1).round(3)
        1.283

    """

    def __init__(self, alpha=1, kappa=1, scale=1, shift=0):
        super(Burr, self).__init__(
            dist=burr(alpha, kappa),
            scale=scale,
            shift=shift,
            repr_args=[alpha, kappa],
        )



"""
Generate Chebyshev pseudo-random samples.

Example usage
-------------

Basic usage::

    >>> distribution = chaospy.Uniform(0, 1)
    >>> samples = distribution.sample(2, rule="chebyshev")
    >>> samples.round(4)
    array([0.25, 0.75])
    >>> samples = distribution.sample(5, rule="chebyshev")
    >>> samples.round(4)
    array([0.067, 0.25 , 0.5  , 0.75 , 0.933])

Certain orders are nested::

    >>> samples = distribution.sample(3, rule="chebyshev")
    >>> samples.round(4)
    array([0.1464, 0.5   , 0.8536])
    >>> samples = distribution.sample(7, rule="chebyshev")
    >>> samples.round(4)
    array([0.0381, 0.1464, 0.3087, 0.5   , 0.6913, 0.8536, 0.9619])

Create nested samples directly with the dedicated function::

    >>> samples = distribution.sample(2, rule="nested_chebyshev")
    >>> samples.round(4)
    array([0.1464, 0.5   , 0.8536])
    >>> samples = distribution.sample(3, rule="nested_chebyshev")
    >>> samples.round(4)
    array([0.0381, 0.1464, 0.3087, 0.5   , 0.6913, 0.8536, 0.9619])

Multivariate usage::

    >>> distribution = chaospy.J(chaospy.Uniform(0, 1), chaospy.Uniform(0, 1))
    >>> samples = distribution.sample(2, rule="chebyshev")
    >>> samples.round(4)
    array([[0.25, 0.25, 0.75, 0.75],
           [0.25, 0.75, 0.25, 0.75]])

"""
import numpy
import chaospy


def create_chebyshev_samples(order, dim=1):
    """
    Chebyshev sampling function.

    Args:
        order (int):
            The number of samples to create along each axis.
        dim (int):
            The number of dimensions to create samples for.

    Returns:
        samples following Chebyshev sampling scheme mapped to the
        ``[0, 1]^dim`` hyper-cube and ``shape == (dim, order)``.
    """
    x_data = .5*numpy.cos(numpy.arange(order, 0, -1)*numpy.pi/(order+1)) + .5
    x_data = chaospy.quadrature.combine([x_data]*dim)
    return x_data.T


def create_nested_chebyshev_samples(order, dim=1):
    """
    Nested Chebyshev sampling function.

    Args:
        order (int):
            The number of samples to create along each axis.
        dim (int):
            The number of dimensions to create samples for.

    Returns:
        samples following Chebyshev sampling scheme mapped to the
        ``[0, 1]^dim`` hyper-cube and ``shape == (dim, 2**order-1)``.
    """
    return create_chebyshev_samples(order=2**order-1, dim=dim)



"""
Multiplication of distributions.

Example usage
-------------

Distribution multiplied with a constant::

    >>> distribution = chaospy.Uniform(0, 1)*4
    >>> distribution
    Multiply(Uniform(), 4)
    >>> distribution.sample(5).round(4)
    array([2.6144, 0.46  , 3.8011, 1.9288, 3.4899])
    >>> distribution.fwd([1, 2, 3]).round(4)
    array([0.25, 0.5 , 0.75])
    >>> distribution.inv(distribution.fwd([1, 2, 3])).round(4)
    array([1., 2., 3.])
    >>> distribution.pdf([1, 2, 3]).round(4)
    array([0.25, 0.25, 0.25])
    >>> distribution.mom([1, 2, 3]).round(4)
    array([ 2.    ,  5.3333, 16.    ])
    >>> distribution.ttr([1, 2, 3]).round(4)
    array([[2.    , 2.    , 2.    ],
           [1.3333, 1.0667, 1.0286]])

Construct joint multiplication distribution::

    >>> lhs = chaospy.Uniform(-1, 0)
    >>> rhs = chaospy.Uniform(-3, -2)
    >>> multiplication = lhs * rhs
    >>> multiplication
    Multiply(Uniform(lower=-1, upper=0), Uniform(lower=-3, upper=-2))
    >>> joint1 = chaospy.J(lhs, multiplication)
    >>> joint1.lower
    array([-1., -0.])
    >>> joint1.upper
    array([0., 3.])
    >>> joint2 = chaospy.J(rhs, multiplication)
    >>> joint2.lower
    array([-3., -0.])
    >>> joint2.upper
    array([-2.,  3.])
    >>> joint3 = chaospy.J(multiplication, lhs)
    >>> joint3.lower
    array([-0., -1.])
    >>> joint3.upper
    array([3., 0.])
    >>> joint4 = chaospy.J(multiplication, rhs)
    >>> joint4.lower
    array([-0., -3.])
    >>> joint4.upper
    array([ 3., -2.])

Generate random samples::

    >>> joint1.sample(4).round(4)
    array([[-0.7877, -0.9593, -0.6028, -0.7669],
           [ 2.2383,  2.1172,  1.6532,  1.8345]])
    >>> joint2.sample(4).round(4)
    array([[-2.8177, -2.2565, -2.9304, -2.1147],
           [ 2.6843,  2.1011,  1.2174,  0.0613]])

Forward transformations::

    >>> lcorr = numpy.array([-0.9, -0.5, -0.1])
    >>> rcorr = numpy.array([-2.99, -2.5, -2.01])
    >>> joint1.fwd([lcorr, lcorr*rcorr]).round(4)
    array([[0.1 , 0.5 , 0.9 ],
           [0.99, 0.5 , 0.01]])
    >>> joint2.fwd([rcorr, lcorr*rcorr]).round(4)
    array([[0.01, 0.5 , 0.99],
           [0.9 , 0.5 , 0.1 ]])

Inverse transformations::

    >>> joint1.inv(joint1.fwd([lcorr, lcorr*rcorr])).round(4)
    array([[-0.9  , -0.5  , -0.1  ],
           [ 2.691,  1.25 ,  0.201]])
    >>> joint2.inv(joint2.fwd([rcorr, lcorr*rcorr])).round(4)
    array([[-2.99 , -2.5  , -2.01 ],
           [ 2.691,  1.25 ,  0.201]])
"""
import numpy
import chaospy

from ..baseclass import Distribution, OperatorDistribution


class Multiply(OperatorDistribution):
    """Multiplication."""

    def __init__(self, left, right):
        """
        Args:
            left (Distribution, numpy.ndarray):
                Left hand side.
            right (Distribution, numpy.ndarray):
                Right hand side.
        """
        self._cache_upper = {}
        self._cache_lower = {}
        super(Multiply, self).__init__(
            left=left,
            right=right,
            repr_args=[left, right],
        )

    def _lower(self, idx, left, right, cache):
        """
        Distribution lower bounds.

        Example:
            >>> chaospy.Multiply(chaospy.Uniform(-1, 2), -2).lower
            array([-4.])
            >>> chaospy.Multiply(chaospy.Uniform(-1, 1), chaospy.Uniform(1, 2)).lower
            array([-2.])

        """
        left = self._parameters["left"]
        right = self._parameters["right"]

        if isinstance(left, Distribution):
            left_upper = left._get_upper(idx, cache=self._upper_cache)
            left_lower = left._get_lower(idx, cache=self._lower_cache)

            if isinstance(right, Distribution):
                right_upper = right._get_upper(idx, cache=self._upper_cache)
                right_lower = right._get_lower(idx, cache=self._lower_cache)

                out = numpy.min(numpy.broadcast_arrays(
                    left_lower*right_lower, left_lower*right_upper,
                    left_upper*right_lower, left_upper*right_upper,
                ), axis=0).T

            else:
                out = numpy.min([left_lower*right[idx], left_upper*right[idx]], axis=0).T

        else:
            assert isinstance(right, Distribution)
            right_upper = right._get_upper(idx, cache=self._upper_cache)
            right_lower = right._get_lower(idx, cache=self._lower_cache)
            out = numpy.min([left[idx]*right_lower,
                             left[idx]*right_upper], axis=0).T

        return out

    def _upper(self, idx, left, right, cache):
        """
        Distribution upper bounds.

        Example:
            >>> chaospy.Multiply(chaospy.Uniform(-1, 2), -2).upper
            array([2.])
            >>> chaospy.Multiply(chaospy.Uniform(-1, 1), chaospy.Uniform(1, 2)).upper
            array([2.])

        """
        # small hack to deal with sign-flipping boundaries.
        left = self._parameters["left"]
        right = self._parameters["right"]
        if isinstance(left, Distribution):
            left_lower = left._get_lower(idx, cache=self._lower_cache)
            left_upper = left._get_upper(idx, cache=self._upper_cache)

            if isinstance(right, Distribution):
                right_lower = right._get_lower(idx, cache=self._lower_cache)
                right_upper = right._get_upper(idx, cache=self._upper_cache)

                out = numpy.max(numpy.broadcast_arrays(
                    left_lower*right_lower, left_lower*right_upper,
                    left_upper*right_lower, left_upper*right_upper,
                ), axis=0)

            else:
                out = numpy.max([left_lower*right[idx],
                                 left_upper*right[idx]], axis=0)

        else:
            assert isinstance(right, Distribution)
            right_lower = right._get_lower(idx, cache=self._lower_cache)
            right_upper = right._get_upper(idx, cache=self._upper_cache)
            out = numpy.max([left[idx]*right_lower,
                             left[idx]*right_upper], axis=0)

        return out

    def _cdf(self, xloc, idx, left, right, cache):
        if isinstance(left, Distribution):
            left, right = right, left
        left = numpy.broadcast_arrays(left.T, xloc.T)[0].T
        valids = left != 0
        xloc_ = xloc.copy()
        xloc_.T[valids.T] = xloc.T[valids.T]/left.T[valids.T]
        uloc = right._get_fwd(xloc_, idx, cache=cache)
        return numpy.where(left.T >= 0, uloc.T, 1-uloc.T).T

    def _ppf(self, uloc, idx, left, right, cache):
        """
        Point percentile function.

        Example:
            >>> chaospy.Uniform().inv([0.1, 0.2, 0.9])
            array([0.1, 0.2, 0.9])
            >>> Multiply(chaospy.Uniform(), 2).inv([0.1, 0.2, 0.9])
            array([0.2, 0.4, 1.8])
            >>> Multiply(2, chaospy.Uniform()).inv([0.1, 0.2, 0.9])
            array([0.2, 0.4, 1.8])
            >>> dist = chaospy.Multiply(
            ...     [2, 1], chaospy.Iid(chaospy.Uniform(), 2))
            >>> dist.inv([[0.5, 0.6, 0.7], [0.5, 0.6, 0.7]])
            array([[1. , 1.2, 1.4],
                   [0.5, 0.6, 0.7]])
            >>> dist = chaospy.Multiply(chaospy.Iid(chaospy.Uniform(), 2), [1, 2])
            >>> dist.inv([[0.5, 0.6, 0.7], [0.5, 0.6, 0.7]])
            array([[0.5, 0.6, 0.7],
                   [1. , 1.2, 1.4]])

        """
        if isinstance(right, Distribution):
            left, right = right, left
        uloc = numpy.where(numpy.asfarray(right).T > 0, uloc.T, 1-uloc.T).T
        xloc = left._get_inv(uloc, idx, cache=cache)
        xloc = (xloc.T*right.T).T

        return xloc

    def _pdf(self, xloc, idx, left, right, cache):
        """
        Probability density function.

        Example:
            >>> chaospy.Uniform().pdf([-0.5, 0.5, 1.5, 2.5])
            array([0., 1., 0., 0.])
            >>> Multiply(chaospy.Uniform(), 2).pdf([-0.5, 0.5, 1.5, 2.5])
            array([0. , 0.5, 0.5, 0. ])
            >>> Multiply(2, chaospy.Uniform()).pdf([-0.5, 0.5, 1.5, 2.5])
            array([0. , 0.5, 0.5, 0. ])
            >>> dist = chaospy.Multiply([2, 1], chaospy.Iid(chaospy.Uniform(), 2))
            >>> dist.pdf([[0.5, 0.6, 1.5], [0.5, 0.6, 1.5]])
            array([0.5, 0.5, 0. ])
            >>> dist = chaospy.Multiply(chaospy.Iid(chaospy.Uniform(), 2), [1, 2])
            >>> dist.pdf([[0.5, 0.6, 1.5], [0.5, 0.6, 1.5]])
            array([0.5, 0.5, 0. ])

        """
        if isinstance(right, Distribution):
            left, right = right, left
        right = (numpy.asfarray(right).T+numpy.zeros(xloc.shape).T).T
        valids = right != 0
        xloc = xloc.copy()
        xloc.T[valids.T] = xloc.T[valids.T]/right.T[valids.T]
        pdf = left._get_pdf(xloc, idx, cache=cache)
        pdf.T[valids.T] /= right.T[valids.T]
        return numpy.abs(pdf)

    def _mom(self, key, left, right, cache):
        """
        Statistical moments.

        Example:
            >>> chaospy.Uniform().mom([0, 1, 2, 3]).round(4)
            array([1.    , 0.5   , 0.3333, 0.25  ])
            >>> Multiply(chaospy.Uniform(), 2).mom([0, 1, 2, 3]).round(4)
            array([1.    , 1.    , 1.3333, 2.    ])
            >>> Multiply(2, chaospy.Uniform()).mom([0, 1, 2, 3]).round(4)
            array([1.    , 1.    , 1.3333, 2.    ])
            >>> Multiply(chaospy.Uniform(), chaospy.Uniform()).mom([0, 1, 2, 3]).round(4)
            array([1.    , 0.25  , 0.1111, 0.0625])

        """
        del cache
        if isinstance(left, Distribution):
            if chaospy.shares_dependencies(left, right):
                raise chaospy.StochasticallyDependentError(
                    "product of dependent distributions not feasible: "
                    "{} and {}".format(left, right)
                )
            left = left._get_mom(key)
        else:
            left = (numpy.array(left).T**key).T
        if isinstance(right, Distribution):
            right = right._get_mom(key)
        else:
            right = (numpy.array(right).T**key).T
        return numpy.prod(left)*numpy.prod(right)

    def _ttr(self, kloc, idx, left, right, cache):
        """Three terms recurrence coefficients."""
        del cache
        if isinstance(right, Distribution):
            if isinstance(left, Distribution):
                raise chaospy.StochasticallyDependentError(
                    "product of distributions not feasible: "
                    "{} and {}".format(left, right)
                )
            left, right = right, left
        coeff0, coeff1 = left._get_ttr(kloc, idx)
        return coeff0*right, coeff1*right*right



"""Wald distribution."""
import numpy
from scipy import special
import chaospy

from ..baseclass import SimpleDistribution, ShiftScaleDistribution


class wald(SimpleDistribution):
    """Wald distribution."""

    def __init__(self, mu):
        super(wald, self).__init__(dict(mu=mu))

    def _pdf(self, x, mu):
        out = numpy.zeros(x.shape)
        indices = x > 0
        out[indices] = 1.0/numpy.sqrt(2*numpy.pi*x[indices])
        out[indices] *= numpy.exp(-(1-mu*x[indices])**2.0 / (2*x[indices]*mu**2.0))
        return out

    def _cdf(self, x, mu):
        trm1 = 1./mu-x
        trm2 = 1./mu+x
        isqx = numpy.tile(numpy.inf, x.shape)
        indices = x > 0
        isqx[indices] = 1./numpy.sqrt(x[indices])
        out = 1.-special.ndtr(isqx*trm1)
        out -= numpy.exp(2.0/mu)*special.ndtr(-isqx*trm2)
        out = numpy.where(x == numpy.inf, 1, out)
        out = numpy.where(x == -numpy.inf, 0, out)
        return out

    def _lower(self, mu):
        return 0.

    def _upper(self, mu):
        qloc = numpy.repeat(1-1e-12, mu.size)
        out = chaospy.approximate_inverse(
            distribution=self,
            idx=0,
            qloc=qloc,
            parameters=dict(mu=mu),
            bounds=(0., 60+numpy.e**(1./(mu+0.1))),
            tolerance=1e-15,
        )
        return out


class Wald(ShiftScaleDistribution):
    """
    Wald distribution.

    Reciprocal inverse Gaussian distribution.

    Args:
        mu (float, Distribution):
            Mean of the normal distribution
        scale (float, Distribution):
            Scaling parameter
        shift (float, Distribution):
            Location parameter

    Examples:
        >>> distribution = chaospy.Wald(0.5)
        >>> distribution
        Wald(0.5)
        >>> uloc = numpy.linspace(0, 1, 6)
        >>> uloc
        array([0. , 0.2, 0.4, 0.6, 0.8, 1. ])
        >>> xloc = distribution.inv(uloc)
        >>> xloc.round(3)
        array([ 0.   ,  1.416,  2.099,  2.94 ,  4.287, 54.701])
        >>> numpy.allclose(distribution.fwd(xloc), uloc)
        True
        >>> distribution.pdf(xloc).round(3)
        array([0.   , 0.297, 0.275, 0.2  , 0.105, 0.   ])
        >>> distribution.sample(4).round(3)
        array([0.653, 1.904, 1.698, 1.161])

    """

    def __init__(self, mu=1, scale=1, shift=0):
        super(Wald, self).__init__(
            dist=wald(mu),
            scale=scale,
            shift=shift,
            repr_args=[mu],
        )



"""
Uncertainty Quantification Toolbox
==================================

This module contains tools for performing uncertainty quantification of models.
"""
import logging
import os
import pkg_resources

from numpoly import *

import chaospy.chol
import chaospy.descriptives
import chaospy.distributions
import chaospy.orthogonal
import chaospy.spectral
import chaospy.quadrature
import chaospy.saltelli
import chaospy.regression

# To be deprecated
from .basis import basis
from .prange import prange

from chaospy.distributions import *
from chaospy.orthogonal import *
from chaospy.spectral import *
from chaospy.quadrature import *
from chaospy.saltelli import *
from chaospy.descriptives import *
from chaospy.regression import *
from chaospy.external import *

try:
    __version__ = pkg_resources.get_distribution("chaospy").version
except pkg_resources.DistributionNotFound:
    __version__ = None


def configure_logging():
    """Configure logging for Chaospy."""
    logpath = os.environ.get("CHAOSPY_LOGPATH", os.devnull)
    logging.basicConfig(level=logging.DEBUG, filename=logpath, filemode="w")
    streamer = logging.StreamHandler()
    loglevel = logging.DEBUG if os.environ.get("CHAOSPY_DEBUG", "") == "1" else logging.WARNING
    streamer.setLevel(loglevel)

    logger = logging.getLogger("chaospy")
    logger.addHandler(streamer)
    logger = logging.getLogger("numpoly")
    logger.addHandler(streamer)

configure_logging()


class StochasticallyDependentError(ValueError):
    """Error related to stochastically dependent variables."""


class UnsupportedFeature(NotImplementedError):
    """Error when dependencies are not correctly handled."""



"""Mielke's beta-kappa distribution."""
import numpy
from scipy import special

from ..baseclass import SimpleDistribution, ShiftScaleDistribution


class mielke(SimpleDistribution):
    """Mielke's beta-kappa distribution."""

    def __init__(self, k, s):
        super(mielke, self).__init__(dict(k=k, s=s))

    def _pdf(self, x, k, s):
        return k*x**(k-1.)/(1.+x**s)**(1.+k*1./s)

    def _cdf(self, x, k, s):
        return x**k/(1.+x**s)**(k*1./s)

    def _ppf(self, q, k, s):
        qsk = pow(q, s*1./k)
        return pow(qsk/(1.-qsk), 1./s)

    def _lower(self, k, s):
        return 0.

    def _upper(self, k, s):
        qsk = pow(1-1e-10, s*1./k)
        return pow(qsk/(1.-qsk), 1./s)


class Mielke(ShiftScaleDistribution):
    """
    Mielke's beta-kappa distribution

    Args:
        kappa (float, Distribution):
            First shape parameter
        expo (float, Distribution):
            Second shape parameter
        scale (float, Distribution):
            Scaling parameter
        shift (float, Distribution):
            Location parameter

    Examples:
        >>> distribution = chaospy.Mielke(kappa=1.5, expo=15)
        >>> distribution
        Mielke(1.5, 15)
        >>> uloc = numpy.linspace(0, 1, 6)
        >>> uloc
        array([0. , 0.2, 0.4, 0.6, 0.8, 1. ])
        >>> xloc = distribution.inv(uloc)
        >>> xloc.round(3)
        array([0.   , 0.342, 0.543, 0.712, 0.868, 3.981])
        >>> numpy.allclose(distribution.fwd(xloc), uloc)
        True
        >>> distribution.pdf(xloc).round(3)
        array([0.   , 0.877, 1.105, 1.257, 1.234, 0.   ])
        >>> distribution.sample(4).round(3)
        array([0.754, 0.236, 1.028, 0.615])

    """

    def __init__(self, kappa=1, expo=1, scale=1, shift=0):
        super(Mielke, self).__init__(
            dist=mielke(kappa, expo),
            scale=scale,
            shift=shift,
            repr_args=[kappa, expo],
        )



"""
Implementation of the Gill-Murray-Wright modified Cholesky algorithm.

Algorithm 6.5 from page 148 of 'Numerical Optimization' by Jorge Nocedal and
Stephen J. Wright, 1999, 2nd ed.
"""
import numpy


def gill_murray_wright(mat, eps=1e-16):
    """
    Gill-Murray-Wright algorithm for pivoting modified Cholesky decomposition.

    Return ``(perm, lowtri, error)`` such that
    `perm.T*mat*perm = lowtri*lowtri.T` is approximately correct.

    Args:
        mat (numpy.ndarray):
            Must be a non-singular and symmetric matrix
        eps (float):
            Error tolerance used in algorithm.

    Returns:
        (:py:data:typing.Tuple[numpy.ndarray, numpy.ndarray]):
            Permutation matrix used for pivoting and lower triangular factor.

    Examples:
        >>> mat = numpy.array([[4, 2, 1], [2, 6, 3], [1, 3, -.004]])
        >>> perm, lowtri = gill_murray_wright(mat)
        >>> perm
        array([[0, 1, 0],
               [1, 0, 0],
               [0, 0, 1]])
        >>> lowtri.round(4)
        array([[ 2.4495,  0.    ,  0.    ],
               [ 0.8165,  1.8257,  0.    ],
               [ 1.2247, -0.    ,  1.2264]])
        >>> (perm @ lowtri @ lowtri.T @ perm.T).round(4)
        array([[4.   , 2.   , 1.   ],
               [2.   , 6.   , 3.   ],
               [1.   , 3.   , 3.004]])
    """
    mat = numpy.asfarray(mat)
    size = mat.shape[0]

    # Calculate gamma(mat) and xi_(mat).
    gamma = 0.0
    xi_ = 0.0
    for idy in range(size):
        gamma = max(abs(mat[idy, idy]), gamma)
        for idx in range(idy+1, size):
            xi_ = max(abs(mat[idy, idx]), xi_)

    # Calculate delta and beta.
    delta = eps * max(gamma + xi_, 1.0)
    if size == 1:
        beta = numpy.sqrt(max(gamma, eps))
    else:
        beta = numpy.sqrt(max(gamma, xi_ / numpy.sqrt(size*size - 1.0), eps))

    # Initialise data structures.
    mat_a = 1.0 * mat
    mat_r = 0.0 * mat
    perm = numpy.eye(size, dtype=int)

    # Main loop.
    for idx in range(size):

        # Row and column swapping, find the index > idx of the largest
        # idzgonal element.
        idz = idx
        for idy in range(idx+1, size):
            if abs(mat_a[idy, idy]) >= abs(mat_a[idz, idz]):
                idz = idy

        if idz != idx:
            mat_a, mat_r, perm = swap_across(idz, idx, mat_a, mat_r, perm)

        # Calculate a_pred.
        theta_j = 0.0
        if idx < size-1:
            for idy in range(idx+1, size):
                theta_j = max(theta_j, abs(mat_a[idx, idy]))
        a_pred = max(abs(mat_a[idx, idx]), (theta_j/beta)**2, delta)

        # Calculate row idx of r and update a.
        mat_r[idx, idx] = numpy.sqrt(a_pred)
        for idy in range(idx+1, size):
            mat_r[idx, idy] = mat_a[idx, idy] / mat_r[idx, idx]
            for idz in range(idx+1, idy+1):

                # Keep matrix a symmetric:
                mat_a[idy, idz] = mat_a[idz, idy] = \
                    mat_a[idz, idy] - mat_r[idx, idy] * mat_r[idx, idz]

    # The Cholesky factor of mat.
    return perm, mat_r.T


def swap_across(idx, idy, mat_a, mat_r, perm):
    """Interchange row and column idy and idx."""
    # Temporary permutation matrix for swaping 2 rows or columns.
    size = mat_a.shape[0]
    perm_new = numpy.eye(size, dtype=int)

    # Modify the permutation matrix perm by swaping columns.
    perm_row = 1.0*perm[:, idx]
    perm[:, idx] = perm[:, idy]
    perm[:, idy] = perm_row

    # Modify the permutation matrix p by swaping rows (same as
    # columns because p = pT).
    row_p = 1.0 * perm_new[idx]
    perm_new[idx] = perm_new[idy]
    perm_new[idy] = row_p

    # Permute mat_a and r (p = pT).
    mat_a = numpy.dot(perm_new, numpy.dot(mat_a, perm_new))
    mat_r = numpy.dot(mat_r, perm_new)
    return mat_a, mat_r, perm



"""Wrapped Cauchy distribution."""
import numpy
from scipy import special

from ..baseclass import SimpleDistribution, ShiftScaleDistribution


class wrapped_cauchy(SimpleDistribution):
    """Wrapped Cauchy distribution."""

    def __init__(self, c):
        super(wrapped_cauchy, self).__init__(dict(c=c))

    def _pdf(self, x, c):
        return (1.0-c*c)/(2*numpy.pi*(1+c*c-2*c*numpy.cos(x)))

    def _cdf(self, x, c):
        output = 0.0*x
        val = (1.0+c)/(1.0-c)
        c1 = x<numpy.pi
        c2 = 1-c1

        xn = numpy.extract(c2,x)
        if (any(xn)):
            valn = numpy.extract(c2, numpy.ones_like(x)*val)
            xn = 2*numpy.pi - xn
            yn = numpy.tan(xn/2.0)
            on = 1.0-1.0/numpy.pi*numpy.arctan(valn*yn)
            numpy.place(output, c2, on)

        xp = numpy.extract(c1,x)
        if (any(xp)):
            valp = numpy.extract(c1, numpy.ones_like(x)*val)
            yp = numpy.tan(xp/2.0)
            op = 1.0/numpy.pi*numpy.arctan(valp*yp)
            numpy.place(output, c1, op)

        return output

    def _ppf(self, qloc, c):
        val = (1.0-c)/(1.0+c)
        rcq = 2*numpy.arctan(val*numpy.tan(numpy.pi*qloc))
        rcmq = 2*numpy.pi-2*numpy.arctan(val*numpy.tan(numpy.pi*(1-qloc)))
        return numpy.where(qloc < 0.5, rcq, rcmq)

    def _lower(self, c):
        return 0.

    def _upper(self, c):
        return 2*numpy.pi


class WrappedCauchy(ShiftScaleDistribution):
    """
    Wrapped Cauchy distribution

    Args:
        shape (float, Distribution):
            Shape parameter
        scale (float, Distribution):
            Scaling parameter
        shift (float, Distribution):
            Location parameter

    Examples:
        >>> distribution = chaospy.WrappedCauchy(0.5)
        >>> distribution
        WrappedCauchy(0.5)
        >>> uloc = numpy.linspace(0, 1, 6)
        >>> uloc
        array([0. , 0.2, 0.4, 0.6, 0.8, 1. ])
        >>> xloc = distribution.inv(uloc)
        >>> xloc.round(3)
        array([0.   , 0.475, 1.596, 4.687, 5.808, 6.283])
        >>> numpy.allclose(distribution.fwd(xloc), uloc)
        True
        >>> distribution.pdf(xloc).round(3)
        array([0.477, 0.331, 0.094, 0.094, 0.331, 0.477])
        >>> distribution.sample(4).round(3)
        array([5.15 , 0.251, 6.178, 2.809])

    """

    def __init__(self, shape=0.5, scale=1, shift=0):
        super(WrappedCauchy, self).__init__(
            dist=wrapped_cauchy(shape),
            scale=scale,
            shift=shift,
            repr_args=[shape],
        )



"""Kumaraswswamy's double bounded distribution."""
import numpy
from scipy import special

from ..baseclass import SimpleDistribution, LowerUpperDistribution


class kumaraswamy(SimpleDistribution):
    """Kumaraswamy's double bounded distribution."""

    def __init__(self, a=1, b=1):
        # assert numpy.all(a > 0) and numpy.all(b > 0)
        super(kumaraswamy, self).__init__(dict(a=a, b=b))

    def _pdf(self, x, a, b):
        return a*b*x**(a-1)*(1-x**a)**(b-1)

    def _cdf(self, x, a, b):
        return 1-(1-x**a)**b

    def _ppf(self, q, a, b):
        return (1-(1-q)**(1./b))**(1./a)

    def _mom(self, k, a, b):
        return (b*special.gamma(1+k*1./a)*special.gamma(b)/
                special.gamma(1+b+k*1./a))

    def _lower(self, a, b):
        return 0.

    def _upper(self, a, b):
        return 1.


class Kumaraswamy(LowerUpperDistribution):
    """
    Kumaraswamy's double bounded distribution

    Args:
        alpha (float, Distribution):
            First shape parameter, alpha > 0
        beta (float, Distribution):
            Second shape parameter, b > 0
        lower (float, Distribution):
            Lower threshold
        upper (float, Distribution):
            Upper threshold

    Examples:
        >>> distribution = chaospy.Kumaraswamy(1.5, 2.5)
        >>> distribution
        Kumaraswamy(1.5, 2.5)
        >>> uloc = numpy.linspace(0, 1, 6)
        >>> uloc
        array([0. , 0.2, 0.4, 0.6, 0.8, 1. ])
        >>> xloc = distribution.inv(uloc)
        >>> xloc.round(3)
        array([0.   , 0.194, 0.324, 0.455, 0.609, 1.   ])
        >>> numpy.allclose(distribution.fwd(xloc), uloc)
        True
        >>> distribution.pdf(xloc).round(3)
        array([0.   , 1.444, 1.572, 1.46 , 1.114, 0.   ])
        >>> distribution.sample(4).round(3)
        array([0.492, 0.132, 0.788, 0.377])
        >>> distribution.mom(1).round(4)
        0.404

    """

    def __init__(self, alpha, beta, lower=0, upper=1):
        super(Kumaraswamy, self).__init__(
            dist=kumaraswamy(alpha, beta),
            lower=lower,
            upper=upper,
            repr_args=[alpha, beta],
        )



# -*- coding: utf-8 -*-
"""
Fejér proposed two quadrature rules very similar to :ref:`clenshaw_curtis`.
The only difference is that the endpoints are set to zero. That is, Fejér only
used the interior extrema of the Chebyshev polynomials, i.e. the true
stationary points. This makes this a better method for performing quadrature on
infinite intervals, as the evaluation does not contain illegal values.

Example usage
-------------

The first few orders with linear growth rule::

    >>> distribution = chaospy.Uniform(0, 1)
    >>> for order in [0, 1, 2, 3]:
    ...     abscissas, weights = chaospy.generate_quadrature(
    ...         order, distribution, rule="fejer")
    ...     print(order, abscissas.round(3), weights.round(3))
    0 [[0.5]] [1.]
    1 [[0.25 0.75]] [0.5 0.5]
    2 [[0.146 0.5   0.854]] [0.286 0.429 0.286]
    3 [[0.095 0.345 0.655 0.905]] [0.188 0.312 0.312 0.188]

The first few orders with exponential growth rule::

    >>> for order in [0, 1, 2]:  # doctest: +NORMALIZE_WHITESPACE
    ...     abscissas, weights = chaospy.generate_quadrature(
    ...         order, distribution, rule="fejer", growth=True)
    ...     print(order, abscissas.round(2), weights.round(2))
    0 [[0.5]] [1.]
    1 [[0.15 0.5  0.85]] [0.29 0.43 0.29]
    2 [[0.04 0.15 0.31 0.5  0.69 0.85 0.96]]
        [0.07 0.14 0.18 0.2  0.18 0.14 0.07]

Applying the rule using Smolyak sparse grid::

    >>> distribution = chaospy.Iid(chaospy.Uniform(0, 1), 2)
    >>> abscissas, weights = chaospy.generate_quadrature(
    ...     2, distribution, rule="fejer", growth=True, sparse=True)
    >>> abscissas.round(3)
    array([[0.038, 0.146, 0.146, 0.146, 0.309, 0.5  , 0.5  , 0.5  , 0.5  ,
            0.5  , 0.5  , 0.5  , 0.691, 0.854, 0.854, 0.854, 0.962],
           [0.5  , 0.146, 0.5  , 0.854, 0.5  , 0.038, 0.146, 0.309, 0.5  ,
            0.691, 0.854, 0.962, 0.5  , 0.146, 0.5  , 0.854, 0.5  ]])
    >>> weights.round(3)
    array([ 0.074,  0.082, -0.021,  0.082,  0.184,  0.074, -0.021,  0.184,
           -0.273,  0.184, -0.021,  0.074,  0.184,  0.082, -0.021,  0.082,
            0.074])
"""
from __future__ import division
try:
    from functools import lru_cache
except ImportError:
    from functools32 import lru_cache

import numpy
import chaospy

from .combine import combine_quadrature


def quad_fejer(order, domain=(0, 1), growth=False, segments=1):
    """
    Generate the quadrature abscissas and weights in Fejer quadrature.

    Args:
        order (int, numpy.ndarray):
            Quadrature order.
        domain (chaospy.distributions.baseclass.Distribution, numpy.ndarray):
            Either distribution or bounding of interval to integrate over.
        growth (bool):
            If True sets the growth rule for the quadrature rule to only
            include orders that enhances nested samples.
        segments (int):
            Split intervals into N subintervals and create a patched
            quadrature based on the segmented quadrature. Can not be lower than
            `order`. If 0 is provided, default to square root of `order`.
            Nested samples only exist when the number of segments are fixed.

    Returns:
        (numpy.ndarray, numpy.ndarray):
            abscissas:
                The quadrature points for where to evaluate the model function
                with ``abscissas.shape == (len(dist), N)`` where ``N`` is the
                number of samples.
            weights:
                The quadrature weights with ``weights.shape == (N,)``.

    Example:
        >>> abscissas, weights = quad_fejer(3, (0, 1))
        >>> abscissas.round(4)
        array([[0.0955, 0.3455, 0.6545, 0.9045]])
        >>> weights.round(4)
        array([0.1804, 0.2996, 0.2996, 0.1804])
        >>> abscissas, weights = quad_fejer(3, (0, 1), segments=2)
        >>> abscissas.round(4)
        array([[0.125, 0.375, 0.625, 0.875]])
        >>> weights.round(4)
        array([0.2222, 0.2222, 0.2222, 0.2222])
    """
    if isinstance(domain, chaospy.Distribution):
        abscissas, weights = quad_fejer(
            order, (domain.lower, domain.upper), growth)
        weights *= domain.pdf(abscissas).flatten()
        weights /= numpy.sum(weights)
        return abscissas, weights

    order = numpy.asarray(order, dtype=int).flatten()
    lower, upper = numpy.array(domain)
    lower = numpy.asarray(lower).flatten()
    upper = numpy.asarray(upper).flatten()

    dim = max(lower.size, upper.size, order.size)

    order = order*numpy.ones(dim, dtype=int)
    lower = lower*numpy.ones(dim)
    upper = upper*numpy.ones(dim)
    segments = segments*numpy.ones(dim, dtype=int)

    if growth:
        order = numpy.where(order > 0, 2**(order+1)-2, 0)

    abscissas, weights = zip(*[_fejer(order_, segment)
                               for order_, segment in zip(order, segments)])

    return combine_quadrature(abscissas, weights, (lower, upper))


@lru_cache(None)
def _fejer(order, segments=1):
    """Backend method."""
    if segments != 1 and order > 2:
        if not segments:
            segments = int(numpy.sqrt(order))
        assert segments < order, "few samples to distribute than intervals"
        abscissas = []
        weights = []

        nodes = numpy.linspace(0, 1, segments+1)
        for idx, (lower, upper) in enumerate(zip(nodes[:-1], nodes[1:])):
            order_ = order//segments + (idx+1 < (order%segments))
            abscissa, weight = _fejer(order_, segments=1)
            abscissas.extend(abscissa*(upper-lower) + lower)
            weights.extend(weight*(upper-lower))

        assert len(abscissas) == order+1, (len(abscissas), order+1)
        assert len(weights) == order+1
        return numpy.array(abscissas), numpy.array(weights)

    order = int(order)
    if order == 0:
        return numpy.array([.5]), numpy.array([1.])
    order += 2

    theta = (order-numpy.arange(order+1))*numpy.pi/order
    abscissas = 0.5*numpy.cos(theta) + 0.5

    idx, idy = numpy.mgrid[:order+1, :order//2]
    weights = 2*numpy.cos(2*(idy+1)*theta[idx])/(4*idy*(idy+2)+3)
    if order % 2 == 0:
        weights[:, -1] *= 0.5
    weights = (1-numpy.sum(weights, -1)) / order

    abscissas = abscissas[1:-1]
    weights = weights[1:-1]
    assert len(abscissas) == order-1
    return abscissas, weights



"""
Laja quadrature is a newer method for performing quadrature in stochastical
problems. The method is described in a `journal paper`_ by Narayan and Jakeman.

.. _journal paper: https://arxiv.org/pdf/1404.5663.pdf

Example usage
-------------

The first few orders::

    >>> distribution = chaospy.Beta(2, 3)
    >>> for order in [0, 1, 2, 3, 4]:  # doctest: +NORMALIZE_WHITESPACE
    ...     abscissas, weights = chaospy.generate_quadrature(
    ...         order, distribution, rule="leja")
    ...     print(order, abscissas.round(3), weights.round(3))
    0 [[0.4]] [1.]
    1 [[0.11 0.4 ]] [0. 1.]
    2 [[0.11  0.4   0.787]] [0.203 0.644 0.153]
    3 [[0.11  0.4   0.6   0.787]] [0.22  0.574 0.086 0.12 ]
    4 [[0.027 0.11  0.4   0.6   0.787]] [-0.047  0.298  0.499  0.146  0.103]

"""
import numpy
from scipy.optimize import fminbound
import chaospy

from .combine import combine
from .recurrence import analytical_stieljes, discretized_stieltjes


def quad_leja(
        order,
        dist,
        rule="fejer",
        accuracy=100,
        recurrence_algorithm="",
):
    """
    Generate Leja quadrature node.

    Args:
        order (int):
            The order of the quadrature.
        dist (chaospy.distributions.baseclass.Distribution):
            The distribution which density will be used as weight function.
        rule (str):
            In the case of ``lanczos`` or ``stieltjes``, defines the
            proxy-integration scheme.
        accuracy (int):
            In the case ``rule`` is used, defines the quadrature order of the
            scheme used. In practice, must be at least as large as ``order``.
        recurrence_algorithm (str):
            Name of the algorithm used to generate abscissas and weights. If
            omitted, ``analytical`` will be tried first, and ``stieltjes`` used
            if that fails.

    Returns:
        (numpy.ndarray, numpy.ndarray):
            abscissas:
                The quadrature points for where to evaluate the model function
                with ``abscissas.shape == (len(dist), N)`` where ``N`` is the
                number of samples.
            weights:
                The quadrature weights with ``weights.shape == (N,)``.

    Example:
        >>> abscissas, weights = quad_leja(3, chaospy.Normal(0, 1))
        >>> abscissas.round(4)
        array([[-2.7173, -1.4142,  0.    ,  1.7635]])
        >>> weights.round(4)
        array([0.022 , 0.1629, 0.6506, 0.1645])
    """
    if len(dist) > 1:
        if dist.stochastic_depedent:
            raise chaospy.StochasticallyDependentError(
                "Leja quadrature do not supper distribution with dependencies.")
        if isinstance(order, int):
            out = [quad_leja(order, _) for _ in dist]
        else:
            out = [quad_leja(order[_], dist[_]) for _ in range(len(dist))]

        abscissas = [_[0][0] for _ in out]
        weights = [_[1] for _ in out]
        abscissas = combine(abscissas).T
        weights = combine(weights)
        weights = numpy.prod(weights, -1)

        return abscissas, weights

    abscissas = [dist.lower, dist.mom(1).flatten(), dist.upper]
    for _ in range(int(order)):

        def objective(abscissas_):
            """Local objective function."""
            out = -numpy.sqrt(dist.pdf(abscissas_))*numpy.prod(
                numpy.abs(abscissas[1:-1]-abscissas_))
            return out

        def fmin(idx):
            """Bound minimization."""
            try:
                x, fx = fminbound(objective, abscissas[idx], abscissas[idx+1], full_output=1)[:2]
            except UnboundLocalError:
                x = abscissas[idx] + 0.5*(3-5**0.5) * (abscissas[idx+1]-abscissas[idx])
                fx = objective(x)
            return x, fx

        opts, vals = zip(*[fmin(idx) for idx in range(len(abscissas)-1)])
        index = numpy.argmin(vals)
        abscissas.insert(index+1, opts[index])

    abscissas = numpy.asfarray(abscissas).flatten()[1:-1]
    weights = create_weights(
        abscissas, dist, rule, accuracy, recurrence_algorithm)
    abscissas = abscissas.reshape(1, abscissas.size)

    return numpy.asfarray(abscissas), numpy.asfarray(weights).flatten()


def create_weights(
        nodes,
        dist,
        rule="fejer",
        accuracy=100,
        recurrence_algorithm="",
):
    """Create weights for the Laja method."""
    try:
        _, poly, _ = analytical_stieljes(len(nodes)-1, dist)
    except NotImplementedError:
        from .frontend import generate_quadrature
        abscissas, weights = generate_quadrature(
            order=accuracy, dist=dist, rule=rule,
            recurrence_algorithm=recurrence_algorithm,
        )
        _, poly, _ = discretized_stieltjes(len(nodes)-1, abscissas, weights)
    poly = poly.flatten()
    weights = numpy.linalg.inv(poly(nodes))
    return weights[:, 0]



"""Normal (Gaussian) probability distribution."""
import numpy
from scipy import special

from ..baseclass import SimpleDistribution, ShiftScaleDistribution


class normal(SimpleDistribution):
    """Standard normal distribution."""

    def __init__(self):
        super(normal, self).__init__()

    def _lower(self):
        return -8.22

    def _upper(self):
        return 8.22

    def _pdf(self, x):
        return (2*numpy.pi)**(-.5)*numpy.e**(-x**2/2.)

    def _cdf(self, x):
        return special.ndtr(x)

    def _ppf(self, x):
        return special.ndtri(x)

    def _mom(self, k):
        return .5*special.factorial2(k-1)*(1+(-1)**k)

    def _ttr(self, n):
        return 0., 1.*n


class Normal(ShiftScaleDistribution):
    R"""
    Normal (Gaussian) distribution

    Args:
        mu (float, Distribution):
            Mean of the distribution.
        sigma (float, Distribution):
            Standard deviation.  sigma > 0

    Examples:
        >>> distribution = chaospy.Normal(2, 3)
        >>> distribution
        Normal(mu=2, sigma=3)
        >>> uloc = numpy.linspace(0, 1, 6)
        >>> uloc
        array([0. , 0.2, 0.4, 0.6, 0.8, 1. ])
        >>> xloc = distribution.inv(uloc)
        >>> xloc.round(3)
        array([-22.66 ,  -0.525,   1.24 ,   2.76 ,   4.525,  26.66 ])
        >>> numpy.allclose(distribution.fwd(xloc), uloc)
        True
        >>> distribution.pdf(xloc).round(3)
        array([0.   , 0.093, 0.129, 0.129, 0.093, 0.   ])
        >>> distribution.sample(4).round(3)
        array([ 3.185, -1.601,  6.943,  1.866])
        >>> distribution.mom(1).round(3)
        2.0
        >>> distribution.ttr([0, 1, 2, 3]).round(3)
        array([[ 2.,  2.,  2.,  2.],
               [ 0.,  9., 18., 27.]])

    """

    def __init__(self, mu=0, sigma=1):
        super(Normal, self).__init__(dist=normal(), scale=sigma, shift=mu)
        self._repr_args = ["mu=%s" % mu, "sigma=%s" % sigma]



"""Chi distribution."""
import numpy
from scipy import special

from ..baseclass import SimpleDistribution, ShiftScaleDistribution


class chi(SimpleDistribution):
    """Chi distribution."""

    def __init__(self, df=1):
        super(chi, self).__init__(dict(df=df))

    def _pdf(self, x, df):
        return x**(df-1)*numpy.exp(-x*x*.5)/2**(df*.5-1)/special.gamma(df*.5)

    def _cdf(self, x, df):
        return special.gammainc(df*0.5,0.5*x*x)

    def _ppf(self, q, df):
        return numpy.sqrt(2*special.gammaincinv(df*0.5, q))

    def _lower(self, df):
        return numpy.sqrt(2*special.gammaincinv(df*0.5, 1e-12))

    def _upper(self, df):
        return numpy.sqrt(2*special.gammaincinv(df*0.5, 1-1e-12))

    def _mom(self, k, df):
        return 2**(.5*k)*special.gamma(.5*(df+k))/special.gamma(.5*df)


class Chi(ShiftScaleDistribution):
    """
    Chi distribution.

    Args:
        df (float, Distribution):
            Degrees of freedom
        scale (float, Distribution):
            Scaling parameter
        shift (float, Distribution):
            Location parameter

    Examples:
        >>> distribution = chaospy.Chi(1.5)
        >>> distribution
        Chi(1.5)
        >>> uloc = numpy.linspace(0, 1, 6)
        >>> uloc
        array([0. , 0.2, 0.4, 0.6, 0.8, 1. ])
        >>> xloc = distribution.inv(uloc)
        >>> xloc.round(3)
        array([0.   , 0.472, 0.791, 1.127, 1.568, 7.294])
        >>> numpy.allclose(distribution.fwd(xloc), uloc)
        True
        >>> distribution.pdf(xloc).round(3)
        array([0.   , 0.596, 0.631, 0.546, 0.355, 0.   ])
        >>> distribution.sample(4).round(3)
        array([1.229, 0.321, 2.234, 0.924])
        >>> distribution.mom(1).round(3)
        1.046

    """

    def __init__(self, df=1, scale=1, shift=0):
        super(Chi, self).__init__(
            dist=chi(df),
            scale=scale,
            shift=shift,
            repr_args=[df],
        )


class Maxwell(ShiftScaleDistribution):
    """
    Maxwell-Boltzmann distribution
    Chi distribution with 3 degrees of freedom

    Args:
        scale (float, Distribution):
            Scaling parameter
        shift (float, Distribution):
            Location parameter

    Examples:
        >>> distribution = chaospy.Maxwell()
        >>> distribution
        Maxwell()
        >>> uloc = numpy.linspace(0, 1, 6)
        >>> uloc
        array([0. , 0.2, 0.4, 0.6, 0.8, 1. ])
        >>> xloc = distribution.inv(uloc)
        >>> xloc.round(3)
        array([0.   , 1.003, 1.367, 1.716, 2.154, 7.676])
        >>> numpy.allclose(distribution.fwd(xloc), uloc)
        True
        >>> distribution.pdf(xloc).round(3)
        array([0.   , 0.485, 0.586, 0.539, 0.364, 0.   ])
        >>> distribution.sample(4).round(3)
        array([1.819, 0.806, 2.798, 1.507])
        >>> distribution.mom(1).round(3)
        1.596

    """

    def __init__(self, scale=1, shift=0):
        super(Maxwell, self).__init__(
            dist=chi(3),
            scale=scale,
            shift=shift,
            repr_args=[],
        )


class Rayleigh(ShiftScaleDistribution):
    """
    Rayleigh distribution

    Args:
        scale (float, Distribution):
            Scaling parameter
        shift (float, Distribution):
            Location parameter

    Examples:
        >>> distribution = chaospy.Rayleigh()
        >>> distribution
        Rayleigh()
        >>> uloc = numpy.linspace(0, 1, 6)
        >>> uloc
        array([0. , 0.2, 0.4, 0.6, 0.8, 1. ])
        >>> xloc = distribution.inv(uloc)
        >>> xloc.round(3)
        array([0.   , 0.668, 1.011, 1.354, 1.794, 7.434])
        >>> numpy.allclose(distribution.fwd(xloc), uloc)
        True
        >>> distribution.pdf(xloc).round(3)
        array([0.   , 0.534, 0.606, 0.541, 0.359, 0.   ])
        >>> distribution.sample(4).round(3)
        array([1.456, 0.494, 2.45 , 1.147])
        >>> distribution.mom(1).round(3)
        1.253

    """
    def __init__(self, scale=1, shift=0):
        super(Rayleigh, self).__init__(
            dist=chi(2),
            scale=scale,
            shift=shift,
            repr_args=[],
        )



"""Log-gamma distribution."""
import numpy
from scipy import special

from ..baseclass import SimpleDistribution, ShiftScaleDistribution


class log_gamma(SimpleDistribution):
    """Log-gamma distribution."""

    def __init__(self, c):
        super(log_gamma, self).__init__(dict(c=c))

    def _pdf(self, x, c):
        return numpy.exp(c*x-numpy.exp(x)-special.gammaln(c))

    def _cdf(self, x, c):
        return special.gammainc(c, numpy.exp(x))

    def _ppf(self, q, c):
        return numpy.log(special.gammaincinv(c, q))

    def _lower(self, c):
        return numpy.log(special.gammaincinv(c, 1e-15))

    def _upper(self, c):
        return numpy.log(special.gammaincinv(c, 1-1e-15))


class LogGamma(ShiftScaleDistribution):
    """
    Log-gamma distribution

    Args:
        shape (float, Distribution):
            Shape parameter
        scale (float, Distribution):
            Scaling parameter
        shift (float, Distribution):
            Location parameter

    Examples:
        >>> distribution = chaospy.LogGamma(1.5)
        >>> distribution
        LogGamma(1.5)
        >>> uloc = numpy.linspace(0, 1, 6)
        >>> uloc
        array([0. , 0.2, 0.4, 0.6, 0.8, 1. ])
        >>> xloc = distribution.inv(uloc)
        >>> xloc.round(3)
        array([-22.836,  -0.688,  -0.068,   0.387,   0.842,   3.597])
        >>> numpy.allclose(distribution.fwd(xloc), uloc)
        True
        >>> distribution.pdf(xloc).round(3)
        array([0.   , 0.243, 0.4  , 0.462, 0.392, 0.   ])
        >>> distribution.sample(4).round(3)
        array([ 0.503, -1.125,  1.364,  0.128])

    """

    def __init__(self, shape=1, scale=1, shift=0):
        super(LogGamma, self).__init__(
            dist=log_gamma(shape),
            scale=scale,
            shift=shift,
            repr_args=[shape],
        )



"""Folded Cauchy distribution."""
import numpy

from ..baseclass import SimpleDistribution, ShiftScaleDistribution


class folded_cauchy(SimpleDistribution):
    """Folded Cauchy distribution."""

    def __init__(self, c=0):
        super(folded_cauchy, self).__init__(dict(c=c))

    def _pdf(self, x, c):
        return 1./(numpy.pi*(1+(x-c)**2))+1/(numpy.pi*(1+(x+c)**2))

    def _cdf(self, x, c):
        return (numpy.arctan(x-c)+numpy.arctan(x+c))/numpy.pi

    def _lower(self, c):
        return 0.

    def _upper(self, c):
        return 1e+16  # actually infinity


class FoldedCauchy(ShiftScaleDistribution):
    """
    Folded Cauchy distribution.

    Args:
        shape (float, Distribution):
            Shape parameter
        scale (float, Distribution):
            Scaling parameter
        shift (float, Distribution):
            Location parameter

    Examples:
        >>> distribution = chaospy.FoldedCauchy(1.5)
        >>> distribution
        FoldedCauchy(1.5)
        >>> uloc = numpy.linspace(0.1, 0.9, 5)
        >>> uloc
        array([0.1, 0.3, 0.5, 0.7, 0.9])
        >>> xloc = distribution.inv(uloc)
        >>> xloc.round(3)
        array([0.489, 1.217, 1.803, 2.67 , 6.644])
        >>> numpy.allclose(distribution.fwd(xloc), uloc)
        True
        >>> distribution.pdf(xloc).round(3)
        array([0.222, 0.333, 0.318, 0.152, 0.016])
        >>> distribution.sample(4).round(3)
        array([1.929, 8.542, 0.311, 1.414])

    Notes:
        The Cauchy distribution is what is known as a "pathological"
        distribution. It is not only infinitely bound, but heavy tailed
        enough that approximate bounds is also infinite for any reasonable
        approximation. This makes both bounds and moments results in
        non-sensibel results. In the case of folded-Cauchy distribution::

            >>> distribution.upper > 1e10
            array([ True])

    """

    def __init__(self, shape=0, scale=1, shift=0):
        super(FoldedCauchy, self).__init__(
            dist=folded_cauchy(shape),
            scale=scale,
            shift=shift,
            repr_args=[shape],
        )
